@@ -9,8 +9,6 @@ import simplejson
 import time
 import traceback
 
-from spawning import spawning_child
-
 
 KEEP_GOING = True
 RESTART_CONTROLLER = False
@@ -25,13 +23,14 @@ DEFAULTS = {
     'dev': True,
     'host': '',
     'port': 8080,
-    'deadman_timeout': 120,
+    'deadman_timeout': 10,
+    'max_memory': None,
 }
 
 
 def environ():
     env = os.environ.copy()
-    env['PYTHONPATH'] = os.environ.get('PYTHONPATH', '')
+    env['PYTHONPATH'] = ':'.join(sys.path)
     return env
 
 
@@ -39,14 +38,15 @@ def spawn_new_children(sock, factory_qual, args, config):
     num_processes = int(config.get('num_processes', 1))
 
     parent_pid = os.getpid()
-    print "(%s) spawning %s children with %s" % (
-        parent_pid, num_processes, spawning_child.__file__)
+    print "(%s) Spawning starting up: %s io processes, %s worker threads, %s worker processes" % (
+        parent_pid, num_processes, config['threadpool_workers'], config['processpool_workers'])
 
-    print "(%s) serving wsgi with configuration:" % (
-        os.getpid(), )
-    prettyconfig = pprint.pformat(config)
-    for line in prettyconfig.split('\n'):
-        print "(%s)\t%s" % (os.getpid(), line)
+    if args.get('verbose'):
+        print "(%s) serving wsgi with configuration:" % (
+            os.getpid(), )
+        prettyconfig = pprint.pformat(config)
+        for line in prettyconfig.split('\n'):
+            print "(%s)\t%s" % (os.getpid(), line)
 
     dev = args.get('dev', False)
     child_pipes = []
@@ -62,9 +62,10 @@ def spawn_new_children(sock, factory_qual, args, config):
 
         if not child_pid:
             os.close(parent_side)
+            os.chdir(os.path.dirname(__file__))
             command = [
-                sys.executable,
-                spawning_child.__file__,
+                'python',
+                '-mspawning.spawning_child',
                 str(parent_pid),
                 str(sock.fileno()),
                 str(child_side),
@@ -120,8 +121,15 @@ def reap_children():
                     break
     else:
         if result:
-            print "(%s) Child %s died with code %s (%s)." % (
-                os.getpid(), pid, result, errno.errorcode.get(result, '?'))
+            signum = result & 0xFF
+            exitcode = (result >> 8) & 0xFF
+
+            if signum:
+                print "(%s) Child died from signal %s with code %s." % (
+                    os.getpid(), signum, exitcode)
+            else:
+                print "(%s) Child %s died with code %s." % (
+                    os.getpid(), pid, exitcode)
             ## The way the code is set up right now it's easier just to panic and
             ## start new children if one of the children dies in a way we didn't expect.
             ## Would probably be better to give this code access to child_pipes
@@ -182,10 +190,9 @@ def restart_controller(factory_qual, args, sock, panic=False):
             start_delay *= 2
         restart_args['start_delay'] = start_delay
 
-    env = '/usr/bin/env'
-    os.execve(
-        env,
-        [env, 'spawn', '-z', simplejson.dumps(restart_args)],
+    os.execvpe(
+        sys.executable,
+        ['python', '-mspawning.spawning_controller', '-z', simplejson.dumps(restart_args)],
         environ())
     ## Never gets here!
 
@@ -210,10 +217,10 @@ def run_controller(factory_qual, args, sock=None):
             if sock is not None:
                 sock.close()
             base = os.path.split(__file__)[0]
+            os.chdir(base)
             args = [
-                sys.executable,
-                os.path.join(
-                    base, 'reloader_svn.py'),
+                'python',
+                'reloader_svn.py',
                 '--pid=' + str(controller_pid),
                 '--dir=' + base,
             ]
@@ -249,8 +256,21 @@ def run_controller(factory_qual, args, sock=None):
         ## Never gets here!
 
 
+def watch_memory(max_memory):
+    process_group = os.getpgrp()
+    while True:
+        time.sleep(MEMORY_WATCH_INTERVAL)
+        out = commands.getoutput('ps -o rss -g %s' % (process_group, ))
+        if sum(int(x) for x in out.split('\n')[1:]) > max_memory:
+            print "(%s) memory watcher restarting processes! Memory exceeded %s" % (
+                os.getpid(), max_memory)
+            os.kill(int(controller_pid), signal.SIGHUP)
+
+
 def main():
     parser = optparse.OptionParser(description="Spawning is an easy-to-use and flexible wsgi server. It supports graceful restarting so that your site finishes serving any old requests while starting new processes to handle new requests with the new code. For the simplest usage, simply pass the dotted path to your wsgi application: 'spawn my_module.my_wsgi_app'")
+    parser.add_option('-v', '--verbose', dest='verbose', action='store_true', help='Display verbose configuration '
+        'information when starting up or restarting.')
     parser.add_option("-f", "--factory", dest='factory', default='spawning.wsgi_factory.config_factory',
         help="""Dotted path (eg mypackage.mymodule.myfunc) to a callable which takes a dictionary containing the command line arguments and figures out what needs to be done to start the wsgi application. Current valid values are: spawning.wsgi_factory.config_factory, spawning.paste_factory.config_factory, and spawning.django_factory.config_factory. The factory used determines what the required positional command line arguments will be. See the spawning.wsgi_factory module for documentation on how to write a new factory.
         """)
@@ -287,6 +307,21 @@ def main():
         help='When killing an old i/o process because the code has changed, don\'t wait '
         'any longer than the deadman timeout value for the process to gracefully exit. '
         'If all requests have not completed by the deadman timeout, the process will be mercilessly killed.')
+    parser.add_option('-l', '--access-log-file', dest='access_log_file', default=None,
+        help='The file to log access log lines to. If not given, log to stdout. Pass /dev/null to discard logs.')
+    parser.add_option('-c', '--coverage', dest='coverage', action='store_true',
+        help='If given, gather coverage data from the running program and make the '
+            'coverage report available from the /_coverage url. See the figleaf docs '
+            'for more info: http://darcs.idyll.org/~t/projects/figleaf/doc/')
+    parser.add_option('-m', '--max-memory', dest='max_memory', type='int', default=0,
+        help='If given, the maximum amount of memory this instance of Spawning '
+            'is allowed to use. If all of the processes started by this Spawning controller '
+            'use more than this amount of memory, send a SIGHUP to the controller '
+            'to get the children to restart.')
+    parser.add_option('-a', '--max-age', dest='max_age', type='int',
+        help='If given, the maximum amount of time (in seconds) an instance of spawning_child '
+            'is allowed to run. Once this time limit has expired a SIGHUP will be sent to '
+            'spawning_controller, causing it to restart all of the child processes.')
     parser.add_option('-z', '--z-restart-args', dest='restart_args',
         help='For internal use only')
 
@@ -318,8 +353,31 @@ def main():
             ## The old fd is still open however, so we close it so we don't leak.
             os.close(restart_args['fd'])
     else:
+        ## We're starting up for the first time.
+        ## Become a process group leader.
+        os.setpgrp()
+        ## Fork off the thing that watches memory for this process group.
+        controller_pid = os.getpid()
+        if (options.max_memory or options.max_age) and not os.fork():
+            env = environ()
+            from spawning import memory_watcher
+            basedir, cmdname = os.path.split(memory_watcher.__file__)
+            if cmdname.endswith('.pyc'):
+                cmdname = cmdname[:-1]
+
+            os.chdir(basedir)
+            command = [
+                'python',
+                cmdname,
+                '--max-age', str(options.max_age),
+                str(controller_pid),
+                str(options.max_memory)]
+            print "command", command
+            os.execve(sys.executable, command, env)
+
         factory = options.factory
         factory_args = {
+            'verbose': options.verbose,
             'host': options.host,
             'port': options.port,
             'num_processes': options.processes,
@@ -328,6 +386,8 @@ def main():
             'watch': options.watch,
             'dev': not options.release,
             'deadman_timeout': options.deadman_timeout,
+            'access_log_file': options.access_log_file,
+            'coverage': options.coverage,
             'args': positional_args,
         }
 
